from parladata_base_api.storages.utils import ParladataObject, Storage


class Area(ParladataObject):
    keys = ["name"]

    def __init__(self, name, id, is_new) -> None:
        self.id = id
        self.name = name
        self.is_new = is_new


class AreaStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)
        self.areas = {}
        self.storage = core_storage

    def load_data(self) -> None:
        for area in self.parladata_api.areas.get_all():
            self.store_area(area, is_new=False)

    def store_area(self, area, is_new) -> Area:
        temp_area = Area(
            name=area["name"],
            id=area["id"],
            is_new=is_new,
        )
        self.areas[temp_area.get_key()] = temp_area
        return temp_area

    def get_or_add_object(self, data) -> Area:
        if not self.areas:
            self.load_data()

        key = Area.get_key_from_dict(data)
        if key in self.areas.keys():
            area = self.areas[key]
        else:
            area = self.parladata_api.set_area(data)
            area = self.store_area(area, is_new=True)
        return area



import logging

from parladata_base_api.storages.utils import ParladataObject, Storage

logger = logging.getLogger("logger")


class PublicQuestion(ParladataObject):
    keys = ["gov_id"]

    def __init__(self, gov_id: str, id: int, is_new: bool) -> None:
        self.id = id
        self.gov_id = gov_id
        self.is_new = is_new


class PublicAnswer(ParladataObject):
    keys = ["gov_id"]

    def __init__(self, gov_id: str, id: int, is_new: bool) -> None:
        self.id = id
        self.gov_id = gov_id
        self.is_new = is_new


class PublicQuestionStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)
        self.public_questions = {}
        self.public_answers = {}

    def load_data(self) -> None:
        if not self.public_questions:
            for public_question in self.parladata_api.public_person_questions.get_all(
                mandate=self.storage.mandate_id
            ):
                self.store_public_question(public_question, False)
            logger.info(f"laoded was {len(self.public_questions)} public questions")
        if not self.public_answers:
            for public_answer in self.parladata_api.public_person_answers.get_all(
                mandate=self.storage.mandate_id
            ):
                self.store_public_answer(public_answer, False)
            logger.info(f"laoded was {len(self.public_answers)} public answers")

    def store_public_question(
        self, public_question: dict, is_new: bool
    ) -> PublicQuestion:
        temp_question = PublicQuestion(
            gov_id=public_question["gov_id"],
            id=public_question["id"],
            is_new=is_new,
        )
        self.public_questions[temp_question.get_key()] = temp_question
        return temp_question

    def store_public_answer(self, public_answer: dict, is_new: bool) -> PublicAnswer:
        temp_answer = PublicQuestion(
            gov_id=public_answer["gov_id"],
            id=public_answer["id"],
            is_new=is_new,
        )
        self.public_answers[temp_answer.get_key()] = temp_answer
        return temp_answer

    def set_public_question(self, data: dict) -> PublicQuestion:
        if not self.public_questions:
            self.load_data()
        added_question = self.parladata_api.public_person_questions.set(data)
        self.store_public_question(added_question, True)
        return added_question

    def check_if_public_question_is_parsed(self, question: dict) -> bool:
        if not self.public_questions:
            self.load_data()
        key = PublicQuestion.get_key_from_dict(question)
        return key in self.public_questions.keys()

    def get_public_question(self, gov_id: str) -> PublicQuestion:
        if not self.public_questions:
            self.load_data()
        return self.public_questions[gov_id]

    def set_public_answer(self, data: dict) -> PublicAnswer:
        if not self.public_answers:
            self.load_data()
        added_answer = self.parladata_api.public_person_answers.set(data)
        self.store_public_answer(added_answer, True)
        return added_answer

    def check_if_public_answer_is_parsed(self, answer: dict) -> bool:
        if not self.public_answers:
            self.load_data()
        key = PublicAnswer.get_key_from_dict(answer)
        return key in self.public_answers.keys()



import logging

from parladata_base_api.storages.utils import ParladataObject, Storage

logger = logging.getLogger("logger")


class Law(ParladataObject):
    keys = ["epa", "mandate"]

    def __init__(
        self, id, epa, text, status, timestamp, uid, classification, mandate, is_new
    ) -> None:
        self.id = id
        self.epa = epa
        self.text = text
        self.status = status
        self.classification = classification
        self.timestamp = timestamp
        self.uid = uid
        self.mandate = mandate
        self.considerations = []
        self.is_new = is_new

    def get_timestamp_of_latest_consideration(self) -> str:
        if not self.considerations:
            return None
        return max([consideration.timestamp for consideration in self.considerations])


class ProcedurePhase(ParladataObject):
    keys = ["name"]

    def __init__(self, id, name) -> None:
        self.id = id
        self.name = name


class LegislationStatuses(ParladataObject):
    keys = ["name"]

    def __init__(self, id, name) -> None:
        self.id = id
        self.name = name

    @classmethod
    def get_obj(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
        }


class LegislationClassification(ParladataObject):
    keys = ["name"]

    def __init__(self, id, name) -> None:
        self.id = id
        self.name = name


class LegislationConsideration(ParladataObject):
    keys = ["legislation", "session"]

    def __init__(self, id, law, timestamp, procedure_phase, session, is_new) -> None:
        self.id = id
        self.legislation = law
        self.timestamp = timestamp
        self.procedure_phase = procedure_phase
        self.session = session
        self.is_new = is_new


class LegislationStorage(Storage):
    def __init__(self, code_storage) -> None:
        super().__init__(code_storage)

        self.legislation = {}
        self.legislation_by_id = {}
        self.statuses_by_id = {None: None}
        self.legislation_classifications = {}
        self.legislation_statuses = {}
        self.procedure_phases = {}
        self.procedure_phases_by_id = {}
        self.legislation_considerations = {}

    def load_data(self) -> None:
        """
        load legislation if not loaded
        """
        if self.legislation:
            return
        logger.debug("Load legislation")
        for (
            legislation_classification
        ) in self.parladata_api.legislation_classifications.get_all():
            classification = LegislationClassification(
                id=legislation_classification["id"],
                name=legislation_classification["name"],
            )
            self.legislation_classifications[classification.get_key()] = classification

        for procedure_phase in self.parladata_api.procedure_phases.get_all():
            procedure_phase_obj = ProcedurePhase(
                id=procedure_phase["id"], name=procedure_phase["name"]
            )
            self.procedure_phases[procedure_phase_obj.get_key()] = procedure_phase_obj
            self.procedure_phases_by_id[procedure_phase_obj.id] = procedure_phase_obj

        for legislation_status in self.parladata_api.legislation_statuses.get_all():
            status = LegislationStatuses(
                id=legislation_status["id"], name=legislation_status["name"]
            )
            self.statuses_by_id[legislation_status["id"]] = status
            self.legislation_statuses[status.get_key()] = status

        for law in self.parladata_api.legislation.get_all(
            mandate=self.storage.mandate_id
        ):
            self.store_object(law, is_new=False)

        # TODO thik about optimizations per session
        for (
            legislation_consideration
        ) in self.parladata_api.legislation_consideration.get_all(
            legislation__mandate=self.storage.mandate_id
        ):
            self.store_legislation_consideration(
                legislation_consideration, is_new=False
            )

    def store_object(self, law_dict, is_new) -> Law:
        if "status" in law_dict.keys():
            status = self.statuses_by_id[law_dict["status"]]
        else:
            status = self.legislation_statuses["in_procedure"]

        law_obj = Law(
            id=law_dict["id"],
            epa=law_dict["epa"],
            text=law_dict["text"],
            status=status,
            timestamp=law_dict["timestamp"],
            classification=law_dict.get("classification", None),
            uid=law_dict["uid"],
            mandate=law_dict["mandate"],
            is_new=is_new,
        )
        self.legislation[law_obj.get_key()] = law_obj
        self.legislation_by_id[law_obj.id] = law_obj
        return law_obj

    def set_law_status(self, law, status_name) -> Law:
        status = self.legislation_statuses[status_name]
        data = {"status": status.id}
        patched_law = self.parladata_api.legislation.patch(law.id, data)
        return self.store_object(patched_law, is_new=False)

    def set_law_as_enacted(self, epa) -> None:
        in_procedure = self.legislation_statuses["in_procedure"]
        enacted = self.legislation_statuses["enacted"]
        key = self.get_key_from_epa(epa)
        if key in self.legislation.keys():
            law = self.legislation[key]
            if law.status == in_procedure or law.status == None:
                # update status
                data = {"status": enacted.id}
                patched_law = self.parladata_api.legislation.patch(law.id, data)
                self.store_object(patched_law, is_new=False)

    def set_law_as_rejected(self, epa) -> None:
        in_procedure = self.legislation_statuses["in_procedure"]
        rejected = self.legislation_statuses["rejected"]
        key = self.get_key_from_epa(epa)
        if key in self.legislation.keys():
            law = self.legislation[key]
            if law.status == in_procedure or law.status == None:
                # update status
                data = {"status": rejected.id}
                patched_law = self.parladata_api.legislation.patch(law.id, data)
                self.store_object(patched_law, is_new=False)

    def set_law_as_in_procedure(self, epa) -> None:
        in_procedure = self.legislation_statuses["in_procedure"]
        key = self.get_key_from_epa(epa)
        if key in self.legislation.keys():
            law = self.legislation[key]
            # update status
            data = {"status": in_procedure.id}
            patched_law = self.parladata_api.legislation.patch(law.id, data)
            self.store_object(patched_law, is_new=False)

    def store_legislation_consideration(
        self, consideration_dict, is_new
    ) -> LegislationConsideration:
        law = self.legislation_by_id[consideration_dict["legislation"]]
        phase = self.procedure_phases_by_id[consideration_dict["procedure_phase"]]
        consideration = LegislationConsideration(
            id=consideration_dict["id"],
            law=law,
            timestamp=consideration_dict["timestamp"],
            procedure_phase=phase,
            session=consideration_dict["session"],
            is_new=is_new,
        )
        law.considerations.append(consideration)
        self.legislation_considerations[consideration.get_key()] = consideration
        return consideration

    def set_law(self, data) -> Law:
        added_law = self.parladata_api.legislation.set(data)
        law_obj = self.store_object(added_law, is_new=True)
        return law_obj

    def get_law(self, epa) -> Law:
        if not self.legislation_statuses:
            self.load_data()
        return self.legislation.get(epa, None)

    def set_legislation_consideration(self, data) -> LegislationConsideration:
        legislation_consideration = self.parladata_api.legislation_consideration.set(
            data
        )
        legislation_consideration = self.store_legislation_consideration(
            legislation_consideration, is_new=True
        )
        return legislation_consideration

    def patch_law(self, law, data) -> Law:
        patched_law = self.parladata_api.legislation.patch(law.id, data)
        law_obj = self.store_object(patched_law, is_new=False)
        return law_obj

    def get_key_from_epa(self, epa) -> str:
        key = Law.get_key_from_dict({"epa": epa, "mandate": self.storage.mandate_id})
        return key

    def is_law_parsed(self, epa) -> bool:
        if not self.legislation_statuses:
            self.load_data()
        key = self.get_key_from_epa(epa)
        return key in self.legislation.keys()

    def has_law_name(self, epa) -> bool:
        if not self.legislation_statuses:
            self.load_data()
        return epa in self.legislation.keys()

    def get_law_by_epa(self, epa) -> Law:
        if not self.legislation_statuses:
            self.load_data()
        key = self.get_key_from_epa(epa)
        return self.legislation.get(key, None)

    def get_or_add_object(self, law_data) -> object:
        if not self.legislation_statuses:
            self.load_data()
        key = Law.get_key_from_dict(law_data)

        if key in self.legislation.keys():
            return self.legislation[key]
        else:
            law = self.set_law(law_data)
        return law

    def update_or_add_law(self, law_data) -> Law:
        if not self.legislation_statuses:
            self.load_data()
        key = Law.get_key_from_dict(law_data)

        if key in self.legislation.keys():
            law = self.legislation[key]
            if law.text == None or law.text == "" or law.classification == None:
                law = self.patch_law(law, law_data)
        else:
            law = self.set_law(law_data)
        return law

    def get_legislation_status_by_name(self, name) -> int:
        if not self.legislation_statuses:
            self.load_data()
        return self.legislation_statuses[name].id

    def prepare_and_set_legislation_consideration(
        self, legislation_consideration
    ) -> LegislationConsideration:
        legislation_consideration_key = LegislationConsideration.get_key_from_dict(
            legislation_consideration
        )
        if not legislation_consideration_key in self.legislation_considerations.keys():
            legislation_consideration = self.set_legislation_consideration(
                legislation_consideration
            )
        else:
            legislation_consideration = self.legislation_considerations[
                legislation_consideration_key
            ]
        return legislation_consideration

    def get_procedure_phase(self, procedure_phase: dict) -> ProcedurePhase:
        key = ProcedurePhase.get_key_from_dict(procedure_phase)
        if not self.procedure_phases:
            self.load_data()
        return self.procedure_phases.get(key, None)

    def get_legislation_classifications_by_name(self, name) -> int:
        if not self.legislation_statuses:
            self.load_data()
        key = LegislationClassification.get_key_from_dict({"name": name})
        try:
            legislation_classifications = self.legislation_classifications.get(key)
        except:
            print(f"name is not in loaded legislation classifications")
            return
        return legislation_classifications.id



import logging
from collections import defaultdict
from datetime import datetime

from parladata_base_api.storages.utils import ParladataObject, Storage

logger = logging.getLogger("logger")


class OrganizationMembership(ParladataObject):
    keys = ["member", "organization", "mandate"]

    def __init__(
        self,
        member_id,
        organization_id,
        start_time,
        end_time,
        mandate,
        id,
        is_new,
        parladata_api,
    ) -> None:
        self.id = id
        self.member = member_id
        self.organization = organization_id
        self.start_time = start_time
        self.end_time = end_time
        self.mandate = mandate
        self.is_new = is_new
        self.parladata_api = parladata_api

    def set_end_time(self, end_time) -> dict:
        self.end_time = end_time
        self.parladata_api.organization_memberships.patch(
            self.id, {"end_time": end_time}
        )


class OrganizationMembershipStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)
        self.memberships = defaultdict(list)

        self.temporary_data = defaultdict(list)
        self.temporary_roles = defaultdict(list)

        self.active_voters = defaultdict(dict)

        self.first_load = False

    def store_object(self, membership, is_new) -> OrganizationMembership:
        temp_membership = OrganizationMembership(
            member_id=membership["member"],
            organization_id=membership["organization"],
            start_time=membership["start_time"],
            end_time=membership.get("end_time", None),
            mandate=membership["mandate"],
            id=membership["id"],
            is_new=is_new,
            parladata_api=self.parladata_api,
        )
        self.memberships[temp_membership.get_key()].append(temp_membership)

        return temp_membership

    def load_data(self) -> None:
        if not self.memberships:
            for membership in self.parladata_api.organizations_memberships.get_all(
                mandate=self.storage.mandate_id
            ):
                self.store_object(membership, is_new=False)
            logger.debug(f"laoded was {len(self.memberships)} memberships")

        if not self.memberships:
            self.first_load = True

    def get_or_add_object(self, data) -> OrganizationMembership:
        if not self.memberships:
            self.load_data()
        key = OrganizationMembership.get_key_from_dict(data)
        if key in self.memberships.keys():
            memberships = self.memberships[key]
            for membership in memberships:
                if not membership.end_time:
                    return membership

        membership = self.set_membership(data)
        return membership

    def set_membership(self, data) -> OrganizationMembership:
        added_membership = self.parladata_api.organization_memberships.set(data)
        new_membership = self.store_object(added_membership, is_new=True)
        return new_membership

    def check_if_membership_is_parsed(self, membership) -> bool:
        key = OrganizationMembership.get_key_from_dict(membership)
        return key in self.memberships.keys()



import requests
from requests.auth import HTTPBasicAuth

from .api import Api


class PeopleApi(Api):
    endpoint = "people"

    def add_person_parser_name(self, person_id, parser_name) -> dict:
        return self._set_object(
            f"people/{person_id}/add_parser_name", {"parser_name": parser_name}
        )

    def upload_image(self, person_id, image_url) -> dict:
        file_name = f"/tmp/person_{person_id}.jpg"
        response = requests.get(image_url)
        with open(file_name, "wb") as f:
            f.write(response.content)
        files = {"image": open(file_name, "rb")}

        return self._patch_object(person_id, data={}, files=files)


class OrganizationsApi(Api):
    endpoint = "organizations"


class SessionsApi(Api):
    endpoint = "sessions"

    def get_speech_count(self, id) -> int:
        url = f"{self.base_url}/speeches/count/?session={id}"
        data = requests.get(url).json()
        if "count" in data.keys():
            return data["count"]
        else:
            return 0

    def unvalidate_speeches(self, session_id) -> dict:
        return self._set_object({}, custom_endpoint="unvalidate_speeches")


class VotesApi(Api):
    endpoint = "votes"

    def delete_vote_ballots(self, vote_id) -> dict:
        return self._delete_object(vote_id, custom_endpoint=f"delete_ballots")


class MotionsApi(Api):
    endpoint = "motions"


class AgendaItemsApi(Api):
    endpoint = "agenda-items"


class QuestionsApi(Api):
    endpoint = "questions"


class AnswersApi(Api):
    endpoint = "answers"


class PublicPersonQuestionsApi(Api):
    endpoint = "public-person-questions"


class PublicPersonAnswersApi(Api):
    endpoint = "public-person-answers"


class LegislationApi(Api):
    endpoint = "legislation"


class LegislationClassificationsApi(Api):
    endpoint = "legislation-classifications"


class ProceduresApi(Api):
    endpoint = "procedures"


class ProcedurePhasesApi(Api):
    endpoint = "procedure-phases"


class LegislationConsiderationApi(Api):
    endpoint = "legislation-consideration"


class LegislationStatusesApi(Api):
    endpoint = "legislation-status"


class PersonMembershipsApi(Api):
    endpoint = "person-memberships"


class OrganizationsMembershipsApi(Api):
    endpoint = "organization-memberships"


class AreasApi(Api):
    endpoint = "areas"


class SpeechesApi(Api):
    endpoint = "speeches"


class BallotsApi(Api):
    endpoint = "ballots"


class LinksApi(Api):
    endpoint = "links"


class MandatesApi(Api):
    endpoint = "mandates"


class ParladataApi(object):
    def __init__(self, api_url, api_user, api_password):
        self.base_url = api_url
        self.session = requests.Session()
        self.session.auth = HTTPBasicAuth(api_user, api_password)

        self.sessions = SessionsApi(self.session, self.base_url)
        self.people = PeopleApi(self.session, self.base_url)
        self.organizations = OrganizationsApi(self.session, self.base_url)
        self.votes = VotesApi(self.session, self.base_url)
        self.motions = MotionsApi(self.session, self.base_url)
        self.agenda_items = AgendaItemsApi(self.session, self.base_url)
        self.questions = QuestionsApi(self.session, self.base_url)
        self.answers = AnswersApi(self.session, self.base_url)
        self.public_person_questions = PublicPersonQuestionsApi(
            self.session, self.base_url
        )
        self.public_person_answers = PublicPersonAnswersApi(self.session, self.base_url)
        self.legislation = LegislationApi(self.session, self.base_url)
        self.legislation_classifications = LegislationClassificationsApi(
            self.session, self.base_url
        )
        self.procedures = ProceduresApi(self.session, self.base_url)
        self.procedure_phases = ProcedurePhasesApi(self.session, self.base_url)
        self.legislation_consideration = LegislationConsiderationApi(
            self.session, self.base_url
        )
        self.legislation_statuses = LegislationStatusesApi(self.session, self.base_url)
        self.person_memberships = PersonMembershipsApi(self.session, self.base_url)
        self.organizations_memberships = OrganizationsMembershipsApi(
            self.session, self.base_url
        )
        self.areas = AreasApi(self.session, self.base_url)
        self.speeches = SpeechesApi(self.session, self.base_url)
        self.ballots = BallotsApi(self.session, self.base_url)
        self.links = LinksApi(self.session, self.base_url)
        self.mandates = MandatesApi(self.session, self.base_url)



import logging

from parladata_base_api.api.endpoints import ParladataApi
from parladata_base_api.storages.area_storage import AreaStorage
from parladata_base_api.storages.legislation_storage import LegislationStorage
from parladata_base_api.storages.membership_storage import MembershipStorage
from parladata_base_api.storages.organization_membership_storage import (
    OrganizationMembershipStorage,
)
from parladata_base_api.storages.organization_storage import OrganizationStorage
from parladata_base_api.storages.people_storage import PeopleStorage
from parladata_base_api.storages.public_question_storage import PublicQuestionStorage
from parladata_base_api.storages.question_storage import QuestionStorage
from parladata_base_api.storages.session_storage import SessionStorage


class DataStorage(object):
    default_procedure_phase = 1

    def __init__(
        self,
        mandate_id: int,
        mandate_start_time: str,
        main_org_id: int,
        api_url: str,
        api_auth_username: str,
        api_auth_password: str,
    ) -> None:
        self.mandate_start_time = mandate_start_time
        self.mandate_id = mandate_id
        self.main_org_id = main_org_id

        self.parladata_api = ParladataApi(api_url, api_auth_username, api_auth_password)

        logging.info(
            f"Initialize storages for mandate {mandate_id} with start time {mandate_start_time}"
        )
        self.session_storage = SessionStorage(self)
        self.legislation_storage = LegislationStorage(self)
        self.people_storage = PeopleStorage(self)
        self.organization_storage = OrganizationStorage(self)
        self.question_storage = QuestionStorage(self)
        self.public_question_storage = PublicQuestionStorage(self)
        self.membership_storage = MembershipStorage(self)
        self.area_storage = AreaStorage(self)
        self.organization_membership_storage = OrganizationMembershipStorage(self)



from parladata_base_api.storages.utils import ParladataObject, Storage


class Organization(ParladataObject):
    keys = ["parser_names"]

    def __init__(
        self,
        name: str,
        id: int,
        parser_names: str,
        is_new: bool,
        gov_id: str = None,
        classification: str = None,
    ) -> None:
        self.id = id
        self.name = name
        self.parser_names = parser_names
        self.gov_id = gov_id
        self.classification = classification
        self.is_new = is_new
        self.memberships = []
        self.active_memberships_by_member_id = {}

    def __repr__(self):
        return f"<Organization: {self.name} [{self.id}]>"


class OrganizationStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)
        self.organizations = {}
        self.organizations_by_id = {}
        self.organizations_by_gov_id = {}
        self.active_memberships_by_member_id = {}

    def load_data(self) -> None:
        for organization in self.parladata_api.organizations.get_all():
            if not organization["parser_names"]:
                continue
            self.store_object(organization, is_new=False)

    def store_object(self, organization: dict, is_new: bool) -> Organization:
        temp_organization = Organization(
            name=organization["name"],
            parser_names=organization["parser_names"],
            id=organization["id"],
            is_new=is_new,
            gov_id=organization.get("gov_id", None),
            classification=organization.get("classification", None),
        )
        self.organizations[temp_organization.get_key()] = temp_organization
        self.organizations_by_id[organization["id"]] = temp_organization
        if temp_organization.gov_id:
            self.organizations_by_gov_id[temp_organization.gov_id] = temp_organization
        return temp_organization

    # def get_object_by_parsername(self, name: str) -> Organization:
    #     if not self.organizations:
    #         self.load_data()
    #     try:
    #         name = name.lower()
    #     except:
    #         return None
    #     for parser_names in self.organizations.keys():
    #         for parser_name in parser_names.split("|"):
    #             if name == parser_name:
    #                 return self.organizations[parser_names]
    #     return None

    def get_or_add_object(
        self, organization_data: dict, add: bool = True
    ) -> Organization:
        if not self.organizations:
            self.load_data()
        organization = self.get_object_by_parsername(
            "organizations", organization_data["name"]
        )
        if organization:
            return organization
        elif not add:
            return None
        response_data = self.parladata_api.organizations.set(organization_data)
        return self.store_object(response_data, is_new=True)

    def get_organization_by_id(self, id: int) -> Organization:
        if not self.organizations:
            self.load_data()
        return self.organizations_by_id.get(id, None)

    def get_organization_by_gov_id(self, gov_id):
        if not self.organizations:
            self.load_data()
        return self.organizations_by_gov_id.get(gov_id, None)



import logging

from parladata_base_api.storages.agenda_item_storage import AgendaItemStorage
from parladata_base_api.storages.utils import ParladataObject, Storage
from parladata_base_api.storages.vote_storage import VoteStorage

logger = logging.getLogger("logger")


class Session(ParladataObject):
    keys = ["gov_id"]

    def __init__(
        self,
        name: str,
        gov_id: str,
        id: int,
        organizations: list,
        start_time: str | None,
        end_time: str | None,
        mandate: int,
        is_new: bool,
        in_review: bool,
        core_storage,
        parladata_api,
    ) -> None:
        # session members
        self.id = id
        self.name = name
        self.organizations = organizations
        self.count = None
        self.start_time = start_time
        self.end_time = end_time
        self.gov_id = gov_id
        self.is_new = is_new
        self.in_review = in_review
        self.mandate = mandate
        self.storage = core_storage
        self.parladata_api = parladata_api

        # session children
        self.vote_storage = VoteStorage(self.storage, self)
        self.agenda_items_storage = AgendaItemStorage(self.storage, self)

    def __str__(self):
        return f"{self.name} [{self.id}]"

    def get_speech_count(self) -> int:
        if self.count == None:
            self.count = self.parladata_api.sessions.get_speech_count(self.id)
        return self.count

    def unvalidate_speeches(self) -> None:
        self.parladata_api.sessions.unvalidate_speeches(self.id)

    def add_speeches(self, data) -> None:
        chunks = [data[x : x + 50] for x in range(0, len(data), 50)]
        logger.debug(f"Adding {len(chunks)} speech chunks")
        for chunk in chunks:
            self.parladata_api.speeches.set(chunk)

    def update_start_time(self, timestamp) -> None:
        self.parladata_api.sessions.patch(
            self.id, {"start_time": timestamp.isoformat()}
        )
        self.start_time = timestamp.isoformat()

    def update_end_time(self, timestamp) -> None:
        self.parladata_api.sessions.patch(self.id, {"end_time": timestamp.isoformat()})
        self.end_time = timestamp.isoformat()

    def patch_session(self, data) -> None:
        self.parladata_api.sessions.patch(self.id, data)


class SessionStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)

        self.sessions = {}
        self.dz_sessions_by_names = {}
        self.sessions_in_review = []

    def load_data(self):
        for session in self.parladata_api.sessions.get_all(
            mandate=self.storage.mandate_id
        ):
            self.store_object(session, is_new=False)

    def store_object(self, session, is_new) -> Session:
        temp_session = Session(
            name=session["name"],
            gov_id=session["gov_id"],
            id=session["id"],
            organizations=session["organizations"],
            start_time=session["start_time"],
            end_time=session["end_time"],
            mandate=session["mandate"],
            is_new=is_new,
            in_review=session["in_review"],
            core_storage=self.storage,
            parladata_api=self.parladata_api,
        )
        self.sessions[temp_session.get_key()] = temp_session
        self.dz_sessions_by_names[temp_session.name.lower()] = temp_session
        if temp_session.in_review:
            self.sessions_in_review.append(temp_session)
        return temp_session

    def get_or_add_object(self, data: dict) -> Session:
        if not self.sessions:
            self.load_data()
        key = Session.get_key_from_dict(data)
        if key in self.sessions.keys():
            return self.sessions[key]
        else:
            data.update(mandate=self.storage.mandate_id)
            session = self.parladata_api.sessions.set(data)
            return self.store_object(session, is_new=True)

    def get_object_or_none(self, data: dict) -> Session:
        if not self.sessions:
            self.load_data()
        key = Session.get_key_from_dict(data)
        return self.sessions.get(key, None)

    def patch_session(self, session: Session, data: dict) -> None:
        self.parladata_api.sessions.set(session.id, data)

        # remove session from sessions_in_review if setted to in_review=False
        if not data.get("in_review", True):
            self.sessions_in_review.remove(session)

        # add session to sessions_in_review if setted to in_review=True
        if data.get("in_review", False):
            self.sessions_in_review.append(session)

    def is_session_in_review(self, session: Session) -> bool:
        return session in self.sessions_in_review

    def get_session_by_name(self, name: str) -> Session:
        if not self.sessions:
            self.load_data()
        return self.dz_sessions_by_names.get(name.lower(), None)



import logging
from collections import defaultdict
from datetime import datetime, timedelta

from parladata_base_api.storages.utils import ParladataObject, Storage

logger = logging.getLogger("logger")


class Membership(ParladataObject):
    keys = ["member", "organization", "on_behalf_of", "role", "mandate"]

    def __init__(
        self,
        person,
        organization,
        on_behalf_of,
        role,
        start_time,
        end_time,
        mandate,
        id,
        is_new,
        parladata_api,
    ) -> None:
        self.id = id
        self.member = person
        self.organization = organization
        self.on_behalf_of = on_behalf_of
        self.role = role
        self.start_time = start_time
        self.end_time = end_time
        self.mandate = mandate
        self.is_new = is_new
        self.parladata_api = parladata_api

    def set_end_time(self, end_time) -> dict:
        self.end_time = end_time
        self.parladata_api.person_memberships.patch(self.id, {"end_time": end_time})


class MembershipStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)
        self.memberships = defaultdict(list)

        self.temporary_data = defaultdict(list)
        self.temporary_roles = defaultdict(list)

        self.active_voters = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        self.first_load = False

    def store_object(self, membership, is_new) -> Membership:
        person = self.storage.people_storage.get_person_by_id(membership["member"])
        organization = self.storage.organization_storage.get_organization_by_id(
            membership["organization"]
        )
        if membership["on_behalf_of"]:
            on_behalf_of = self.storage.organization_storage.get_organization_by_id(
                membership["on_behalf_of"]
            )
        else:
            on_behalf_of = None
        temp_membership = Membership(
            person=person,
            organization=organization,
            on_behalf_of=on_behalf_of,
            role=membership["role"],
            start_time=membership["start_time"],
            end_time=membership.get("end_time", None),
            mandate=membership["mandate"],
            id=membership["id"],
            is_new=is_new,
            parladata_api=self.parladata_api,
        )
        self.memberships[temp_membership.get_key()].append(temp_membership)

        if not membership.get("end_time", None):
            organization.active_memberships_by_member_id[membership["member"]] = (
                temp_membership
            )

        if (
            membership.get("end_time", None) == None
            or membership["end_time"] > datetime.now().isoformat()
        ) and membership["role"] == "voter":
            self.active_voters[membership["organization"]][membership["member"]][
                membership["on_behalf_of"]
            ].append(temp_membership)

        if person and (not membership.get("end_time", None)):
            person.active_memberships.append(temp_membership)
        if organization:
            organization.memberships.append(temp_membership)
        return temp_membership

    def load_data(self) -> None:
        if not self.memberships:
            for membership in self.parladata_api.person_memberships.get_all(
                mandate=self.storage.mandate_id
            ):
                self.store_object(membership, is_new=False)
            logger.debug(f"laoded was {len(self.memberships)} memberships")

        if not self.memberships:
            self.first_load = True

        if self.first_load:
            self.default_start_time = self.storage.mandate_start_time.isoformat()
        else:
            self.default_start_time = datetime.now().isoformat()

    def get_or_add_object(self, data) -> Membership:
        if not self.memberships:
            self.load_data()
        key = Membership.get_key_from_dict(data)
        if key in self.memberships.keys():
            memberships = self.memberships[key]
            for membership in memberships:
                if not membership.end_time:
                    return membership

        membership = self.set_membership(data)
        return membership

    def set_membership(self, data) -> Membership:
        added_membership = self.parladata_api.person_memberships.set(data)
        new_membership = self.store_object(added_membership, is_new=True)
        return new_membership

    def get_id_if_membership_is_parsed(self, membership) -> bool:
        key = Membership.get_key_from_dict(membership)
        if key in self.memberships.keys():
            return self.memberships[key][0]
        return None

    def get_membership_in_organiaztion(self, person, organization_id) -> Membership:
        for membership in person.active_memberships:
            if membership.organization.id == organization_id:
                return membership

    def refresh_per_person_memberships(self) -> None:
        """ """
        self.keep_memebrship_ids = []
        memberships_to_end = []

        for person_memberships in self.temporary_data.values():
            party_membership = person_memberships.get("party", None)
            if party_membership:
                party_membership = party_membership[0]
                print(party_membership)
                self.membership_proccessing(party_membership, is_party_membership=True)

            for membership in person_memberships.get("commitee", []):
                self.membership_proccessing(party_membership, is_party_membership=True)

                # membership["on_behalf_of"] = party["organization"]
                # self.membership_storage.temporary_data[
                #     membership["organization"].id
                # ].append(membership)

    def membership_proccessing(
        self,
        single_org_membership,
        is_party_membership: bool = False,
    ) -> None:
        if membership := self.get_id_if_membership_is_parsed(single_org_membership):
            self.keep_memebrship_ids.append(membership.id)
            return
        # get start&end time for the membership
        if start_time := single_org_membership.get("start_time"):
            pass
        else:
            start_time = self.default_start_time

        if end_time := single_org_membership.get("end_time"):
            pass
        else:
            end_time = (
                datetime.fromisoformat(self.default_start_time) - timedelta(seconds=1)
            ).isoformat()

        self.end_time = end_time

        need_to_add_voter_membership = single_org_membership["is_voter"]

        organization = single_org_membership["organization"]
        on_behalf_of = single_org_membership["on_behalf_of"]

        member = single_org_membership["member"]
        role = single_org_membership.get("role", "member")

        if self.temporary_roles:
            if is_party_membership:
                role = self.get_members_role_in_organization(member.id, on_behalf_of.id)
            else:
                role = self.get_members_role_in_organization(member.id, organization.id)
        logger.debug(role)

        # check if user changed party role
        if (
            organization
            and on_behalf_of
            and organization.id == self.storage.main_org_id
        ):
            existing_party_membrships = self.get_membership_in_organiaztion(
                member, on_behalf_of.id
            )
            if existing_party_membrships:
                if existing_party_membrships.role != role:
                    # if user changed party role
                    existing_party_membrships.set_end_time(self.end_time)
                    need_to_add_voter_membership = False

        if on_behalf_of:
            if is_party_membership:
                org_id = on_behalf_of.id
                on_behalf_of_id = None
            else:
                org_id = organization.id
                on_behalf_of_id = on_behalf_of.id
            stored_membership = self.get_or_add_object(
                {
                    "member": member.id,
                    "organization": org_id,
                    "role": role,
                    "start_time": start_time,
                    "mandate": self.storage.mandate_id,
                    "on_behalf_of": None,
                }
            )
            self.keep_memebrship_ids.append(stored_membership.id)
        else:
            if is_party_membership:
                org_id = self.storage.main_org_id

        if need_to_add_voter_membership:
            stored_membership = self.get_or_add_object(
                {
                    "member": member.id,
                    "organization": organization.id,
                    "role": "voter",
                    "start_time": start_time,
                    "mandate": self.storage.mandate_id,
                    "on_behalf_of": on_behalf_of.id if on_behalf_of else None,
                }
            )
            self.keep_memebrship_ids.append(stored_membership.id)
            if stored_membership.is_new:
                self.fix_user_membership(member, org_id, on_behalf_of)

    def refresh_memberships(self) -> None:
        """ """
        self.keep_memebrship_ids = []
        memberships_to_end = []

        for org_id, org_data in self.temporary_data.items():
            for single_org_membership in org_data:
                # check if membership is already parsed
                if membership := self.get_id_if_membership_is_parsed(
                    single_org_membership
                ):
                    self.keep_memebrship_ids.append(membership.id)
                    continue
                # get start&end time for the membership
                if start_time := single_org_membership.get("start_time"):
                    pass
                else:
                    start_time = self.default_start_time

                if end_time := single_org_membership.get("end_time"):
                    pass
                else:
                    end_time = (
                        datetime.fromisoformat(self.default_start_time)
                        - timedelta(seconds=1)
                    ).isoformat()

                self.end_time = end_time

                need_to_add_voter_membership = single_org_membership["is_voter"]

                logger.debug(single_org_membership)
                # get organization id
                if single_org_membership["organization"]:
                    organization = single_org_membership["organization"]
                else:
                    organization = None

                member = single_org_membership["member"]
                role = single_org_membership.get("role", "member")
                if self.temporary_roles:
                    # override role if it's parserd from separate page and stored
                    # in temporary_roles
                    role = self.get_members_role_in_organization(
                        member.id, organization.id
                    )
                logger.debug(role)

                # check if user changed party role
                if organization and org_id == self.storage.main_org_id:
                    existing_party_membrships = self.get_membership_in_organiaztion(
                        member, organization.id
                    )
                    if existing_party_membrships:
                        if existing_party_membrships.role != role:
                            # if user changed party role
                            existing_party_membrships.set_end_time(self.end_time)
                            need_to_add_voter_membership = False

                if not organization:
                    # if membership is None then not add party membership
                    party_id = None

                elif organization and org_id == self.storage.main_org_id:
                    # if current memberships if in main organization then add membership
                    # to party
                    party_id = organization.id
                else:
                    # membership represent committee like membership
                    party = single_org_membership.get("on_behalf_of", None)
                    if party:
                        party_id = party.id
                    else:
                        party_id = None
                    org_id = organization.id

                if party_id:
                    stored_membership = self.get_or_add_object(
                        {
                            "member": member.id,
                            "organization": party_id,
                            "role": role,
                            "start_time": start_time,
                            "mandate": self.storage.mandate_id,
                            "on_behalf_of": None,
                        }
                    )
                    self.keep_memebrship_ids.append(stored_membership.id)

                if need_to_add_voter_membership:
                    stored_membership = self.get_or_add_object(
                        {
                            "member": member.id,
                            "organization": org_id,
                            "role": "voter",
                            "start_time": start_time,
                            "mandate": self.storage.mandate_id,
                            "on_behalf_of": party_id,
                        }
                    )
                    self.keep_memebrship_ids.append(stored_membership.id)
                    if stored_membership.is_new:
                        self.fix_user_membership(member, org_id, organization)

        logger.debug(f"changed mambership {memberships_to_end}")

    # def end_old_memberships_after_parsing(self) -> None:
    #     for org_id, voters in self.active_voters.items():
    #         for person_id, memberships in voters.items():
    #             for membership in memberships:
    #                 if membership.id not in self.keep_memebrship_ids:

    def fix_user_membership(self, person, organization_id, on_behalf_of) -> None:
        """
        This method is used to end old membership if person changed club
        and has more than one membership in the same organization (main org / commitee)

        :param person: Person object
        :param organization_id: Organization id
        :param on_behalf_of: Organization object
        """
        person_voter_org_dict = self.active_voters.get(organization_id).get(person.id)
        if person_voter_org_dict and len(person_voter_org_dict) > 1:
            # person has change a club
            for org_id, org_memberships in person_voter_org_dict.items():
                if org_id != on_behalf_of:
                    for m in org_memberships:
                        m.set_end_time(self.end_time)
                        m.member.active_memberships.remove(m)
                        if organization_id == self.storage.main_org_id:
                            for pm in m.member.active_memberships:
                                if pm.organization == m.on_behalf_of and pm.role in [
                                    "member",
                                    "president",
                                    "deputy",
                                ]:
                                    pm.set_end_time(self.end_time)
                                    pm.member.active_memberships.remove(pm)

    def get_all_active_persons_memberships(self, person_id) -> list:
        return [
            membership
            for membership in self.storage.people_storage.get_person_by_id(
                person_id
            ).memberships
            if not membership.end_time
        ]

    def count_active_voter_membership(self, person_id) -> int:
        person = self.storage.people_storage.get_person_by_id(person_id)
        count = 0
        for membership in person.active_memberships:
            if (
                membership.organization == int(self.storage.main_org_id)
                and not membership.end_time
                and membership.role == "voter"
            ):
                count += 1
        return count

    def get_members_role_in_organization(self, member_id, organization) -> str:
        role = "member"
        for person_role in self.temporary_roles[organization]:
            if str(member_id) == str(person_role["member"].id):
                return person_role["role"]
        return role

    def get_members_organization_from_roles(self, member_id) -> int:
        """
        This is used for users without profiles
        """
        role = None
        for org_id, members in self.temporary_roles.items():
            for person_role in members:
                if str(member_id) == str(person_role["member"].id):
                    role = person_role["role"]
                    return org_id
        return None

    # def count_active_voter_membership(self) -> int:
    #     count = 0
    #     for membership in self.memberships:
    #         if (
    #             membership.organization_id == int(self.storage.main_org_id)
    #             and not membership.end_time
    #             and membership.role == "voter"
    #         ):
    #             count += 1
    #     return count



from parladata_base_api.storages.utils import ParladataObject, Storage


class AgendaItem(ParladataObject):
    key = ["name", "session"]

    def __init__(self, name, id, datetime, session, is_new) -> None:
        self.id = id
        self.name = name
        self.session = session
        self.datetime = datetime
        self.is_new = is_new

    def set_link(self, url, name):
        self.parladata_api.links.set({"url": url, "name": name, "agenda_item": self.id})


class AgendaItemStorage(Storage):
    def __init__(self, core_storage, session) -> None:
        super().__init__(core_storage)
        self.agenda_items = {}
        self.session = session

    def load_data(self) -> None:
        for agenda_item in self.parladata_api.agenda_items.get_all(
            session=self.session.id
        ):
            self.store_object(agenda_item, is_new=False)

    def store_object(self, agenda_item, is_new) -> AgendaItem:
        temp_agenda_item = AgendaItem(
            name=agenda_item["name"],
            datetime=agenda_item["datetime"],
            id=agenda_item["id"],
            session=self.session,
            is_new=is_new,
        )
        self.agenda_items[temp_agenda_item.get_key()] = temp_agenda_item
        return temp_agenda_item

    def get_or_add_object(self, data) -> AgendaItem:
        if not self.agenda_items:
            self.load_data()

        key = AgendaItem.get_key_from_dict(data)
        if key in self.agenda_items.keys():
            agenda_item = self.agenda_items[key]
        else:
            added_agenda_item = self.parladata_api.agenda_items.set(data)
            agenda_item = self.store_object(added_agenda_item, is_new=True)
        return agenda_item



from parladata_base_api.storages.utils import ParladataObject, Storage


class Motion(ParladataObject):
    keys = ["text", "datetime"]

    def __init__(
        self,
        id: int,
        text: str,
        title: str,
        session: int,
        datetime: str,
        gov_id: str,
        is_new: bool,
        parladata_api,
        core_storage,
    ) -> None:
        self.id = id
        self.text = text
        self.title = title
        self.session = session
        self.datetime = datetime
        self.gov_id = gov_id
        self.is_new = is_new
        self.vote = None
        self.storage = core_storage
        self.parladata_api = parladata_api

    def patch(self, data: dict) -> dict:
        self.parladata_api.motions.patch(self.id, data)


class Vote(ParladataObject):
    keys = ["name", "timestamp"]

    def __init__(
        self,
        id: int,
        name: str,
        timestamp: str,
        has_anonymous_ballots: bool,
        is_new: bool,
        core_storage,
        parladata_api,
    ) -> None:
        self.id = id
        self.name = name
        self.timestamp = timestamp
        self.has_anonymous_ballots = has_anonymous_ballots
        self.is_new = is_new
        self.storage = core_storage
        self.parladata_api = parladata_api

    def delete_ballots(self):
        self.parladata_api.votes.delete_vote_ballots(self.id)

    def patch(self, data: dict):
        self.parladata_api.votes.patch(self.id, data)


class VoteStorage(Storage):
    def __init__(self, core_storage, session) -> None:
        super().__init__(core_storage)
        self.motions = {}
        self.anonymous_motions = None

        self.session = session

    def load_data(self) -> None:
        votes_by_motion_id = {
            vote["motion"]: vote
            for vote in self.parladata_api.votes.get_all(
                motion__session=self.session.id
            )
        }
        for motion in self.parladata_api.motions.get_all(session=self.session.id):
            temp_motion = self.store_motion(motion, False)
            vote = votes_by_motion_id[temp_motion.id]
            self.store_vote(vote, temp_motion, False)

    def store_motion(self, data: dict, is_new: bool) -> Motion:
        motion = Motion(
            text=data["text"],
            title=data["title"],
            id=data["id"],
            session=data["session"],
            gov_id=data["gov_id"],
            datetime=data["datetime"],
            is_new=is_new,
            core_storage=self.storage,
            parladata_api=self.parladata_api,
        )
        self.motions[motion.get_key()] = motion
        return motion

    def store_vote(self, data: dict, motion: Motion, is_new: bool) -> Vote:
        vote = Vote(
            name=data["name"],
            id=data["id"],
            timestamp=data["timestamp"],
            has_anonymous_ballots=data["has_anonymous_ballots"],
            is_new=is_new,
            core_storage=self.storage,
            parladata_api=self.parladata_api,
        )
        motion.vote = vote
        return vote

    def set_ballots(self, data: dict) -> dict:
        self.parladata_api.ballots.set(data)

    def set_motion(self, data: dict) -> Motion:
        added_motion = self.parladata_api.motions.set(data)
        return self.store_motion(added_motion, True)

    def set_vote(self, data: dict, motion: Motion) -> Vote:
        added_vote = self.parladata_api.votes.set(data)
        return self.store_vote(added_vote, motion, True)

    def get_or_add_object(self, data: dict) -> Motion:
        if not self.motions:
            self.load_data()
        if self.check_if_motion_is_parsed(data):
            key = Motion.get_key_from_dict(data)
            return self.motions[key]
        else:
            motion = self.set_motion(data)
            data["timestamp"] = data["datetime"]
            del data["datetime"]
            data["motion"] = motion.id
            data["name"] = data["title"]
            self.set_vote(data, motion)
            return motion

    def check_if_motion_is_parsed(self, motion: dict) -> bool:
        if not self.motions:
            self.load_data()
        key = Motion.get_key_from_dict(motion)
        return self.motions.get(key, None)



class Storage(object):
    def __init__(self, core_storage) -> None:
        self.storage = core_storage
        self.parladata_api = core_storage.parladata_api

    def get_or_add_object(self, data) -> object:
        raise NotImplementedError

    def store_object(self, data) -> object:
        raise NotImplementedError

    def load_data(self) -> None:
        raise NotImplementedError

    def get_object_by_parsername(self, object_type: str, name: str) -> object:
        """ """
        name = name.lower()
        for parser_names in getattr(self, object_type).keys():
            for parser_name in parser_names.split("|"):
                if name == parser_name:
                    return getattr(self, object_type)[parser_names]
        return None

    def get_object_by_parsername_compare_genitiv(
        self, object_type: str, name: str
    ) -> object:
        cutted_name = [word[:-2] for word in name.lower().split(" ")]
        for parser_names in getattr(self, object_type).keys():
            for parser_name in parser_names.split("|"):
                cutted_parser_name = [
                    word[:-2] for word in parser_name.lower().split(" ")
                ]
                if len(cutted_parser_name) != len(cutted_name):
                    continue
                result = []
                for i, parted_parser_name in enumerate(cutted_parser_name):
                    result.append(parted_parser_name in cutted_name[i])
                if result and all(result):
                    return getattr(self, object_type)[parser_names]
        return None


class ParladataObject(object):
    keys = ["gov_id"]

    def get_key(self) -> str:
        return "_".join([self._parse_key(k, None) for k in self.keys])

    @classmethod
    def get_key_from_dict(ctl, data) -> str:
        return "_".join([ctl._parse_key(ctl, k, data) for k in ctl.keys])

    @classmethod
    def _parse_value(ctl, value: any) -> str:
        if isinstance(value, str):
            return value.strip().lower()
        elif isinstance(value, int):
            return str(value)
        elif isinstance(value, list):
            value.sort()
            return "-".join([ctl._parse_value(v) for v in value])
        elif isinstance(value, object):
            return str(value.id) if value else "-"
        elif isinstance(value, type(None)):
            return "None"
        else:
            return "-"

    def _parse_key(self, key: str, data: any = None) -> str:
        if isinstance(data, dict):
            value = data[key]
        else:
            value = getattr(self, key)

        return self._parse_value(value)



from endpoints import (
    AgendaItemsApi,
    AnswersApi,
    AreasApi,
    BallotsApi,
    LegislationApi,
    LegislationClassificationsApi,
    LegislationConsiderationApi,
    LegislationStatusesApi,
    LinksApi,
    MandatesApi,
    MotionsApi,
    OrganizationsApi,
    OrganizationsMembershipsApi,
    PeopleApi,
    PersonMembershipsApi,
    ProcedurePhasesApi,
    ProceduresApi,
    PublicPersonAnswersApi,
    PublicPersonQuestionsApi,
    QuestionsApi,
    SessionsApi,
    SpeechesApi,
    VotesApi,
)



import logging

from requests.auth import HTTPBasicAuth

logger = logging.getLogger("logger")


class Api(object):
    def __init__(self, resquests_session, base_url):
        self.session = resquests_session
        self.base_url = base_url
        endpoint = "base"

    def _get_data_from_pager_api_gen(self, url, limit=300):
        end = False
        page = 1
        if "?" in url:
            url = url + f"&limit={limit}"
        else:
            url = url + f"?limit={limit}"
        while url:
            response = self.session.get(url)
            if response.status_code != 200:
                logger.warning(response.content)
                logger.warning(url)
            data = response.json()
            yield data["results"]
            url = data["next"]

    def _get_objects(self, limit=300, *args, **kwargs):
        url = f"{self.base_url}/{self.endpoint}"

        args = "&".join([f"{key}={value}" for key, value in kwargs.items()])
        if args:
            if "?" in url:
                url = url + "&" + args
            else:
                url = url + "?" + args

        return [
            obj
            for page in self._get_data_from_pager_api_gen(url, limit)
            for obj in page
        ]

    def _get_object(self, object_id, custom_endpoint=None):
        url = f"{self.base_url}/{self.endpoint}/{object_id}/" + (
            f"{custom_endpoint}/" if custom_endpoint else ""
        )
        response = self.session.get(url)
        if response.status_code > 299:
            logger.warning(response.content)
            logger.warning(url)
        return response.json()

    def _set_object(self, data, custom_endpoint=None):
        url = f"{self.base_url}/{self.endpoint}/" + (
            f"{custom_endpoint}/" if custom_endpoint else ""
        )
        response = self.session.post(url, json=data)
        if response.status_code > 299:
            logger.warning(response.content)
            logger.warning(url)
        return response.json()

    def _patch_object(self, object_id, data, custom_endpoint=None, files=None):
        url = f"{self.base_url}/{self.endpoint}/{object_id}/" + (
            f"{custom_endpoint}/" if custom_endpoint else ""
        )
        if files:
            response = self.session.patch(url, files=files)
        else:
            response = self.session.patch(url, json=data)
        if response.status_code > 299:
            logger.warning(response.content)
            logger.warning(url)
        return response.json()

    def _delete_object(self, object_id, custom_endpoint=None):
        url = f"{self.base_url}/{self.endpoint}/{object_id}/" + (
            f"{custom_endpoint}/" if custom_endpoint else ""
        )
        response = self.session.delete(url)
        if response.status_code > 299:
            logger.warning(response.content)
            logger.warning(url)
        return response.json()

    def get_all(self, limit=300, *args, **kwargs) -> list:
        return self._get_objects(limit, *args, **kwargs)

    def get(self, person_id) -> dict:
        return self._get_object(person_id)

    def set(self, data) -> dict:
        return self._set_object(data)

    def patch(self, object_id, data, files=None) -> dict:
        return self._patch_object(object_id, data, files=files)

    def delete(self, person_id) -> dict:
        return self._delete_object(person_id)



import re

from parladata_base_api.storages.utils import ParladataObject, Storage


class Person(ParladataObject):
    keys = ["parser_names"]

    def __init__(
        self, name: str, id: int, parser_names: str, is_new: bool, parladata_api
    ) -> None:
        self.id = id
        self.name = name
        self.parser_names = parser_names
        self.is_new = is_new
        self.active_memberships = []
        self.parladata_api = parladata_api

    def save_image(self, image_url: str) -> None:
        self.parladata_api.people.upload_image(self.id, image_url)

    def add_parser_name(self, parser_name: str) -> None:
        data = self.parladata_api.people.add_person_parser_name(self.id, parser_name)
        self.parser_names = data["parser_names"]

    def __repr__(self):
        return f"<Person {self.name} [{self.id}]>"


class PeopleStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)
        self.people = {}
        self.people_by_id = {}
        self.storage = core_storage

    def load_data(self) -> None:
        for person in self.parladata_api.people.get_all():
            self.store_object(person, is_new=False)

    def store_object(self, person: dict, is_new: bool) -> Person:
        temp_person = Person(
            name=person["name"],
            parser_names=person["parser_names"],
            id=person["id"],
            is_new=is_new,
            parladata_api=self.parladata_api,
        )
        self.people[temp_person.get_key()] = temp_person
        self.people_by_id[person["id"]] = temp_person
        return temp_person

    # def get_object_by_parsername(self, name: str) -> Person:
    #     if not self.people:
    #         self.load_data()
    #     try:
    #         name = name.lower()
    #     except:
    #         return None
    #     for parser_names in self.people.keys():
    #         for parser_name in parser_names.split("|"):
    #             if name == parser_name:
    #                 return self.people[parser_names]
    #     return None

    # def get_object_by_parsername_compare_rodilnik(self, object_type, name):
    #     """
    #     """
    #     cutted_name = [word[:-2] for word in name.lower().split(' ')]
    #     for parser_names in self.people.keys():
    #         for parser_name in parser_names.split('|'):
    #             cutted_parser_name = [word[:-2] for word in parser_name.lower().split(' ')]
    #             if len(cutted_parser_name) != len(cutted_name):
    #                 continue
    #             result = []
    #             for i, parted_parser_name in enumerate(cutted_parser_name):
    #                 result.append( parted_parser_name in cutted_name[i] )
    #             if result and all(result):
    #                 return getattr(self, object_type)[parser_names]
    #     return None

    def get_or_add_object(
        self, person_data: dict, add: bool = True, name_type: str = "normal"
    ) -> Person:
        if not self.people:
            self.load_data()
        name = person_data["name"]
        prefix, name = self.get_prefix(name)
        if name_type == "genitive":
            person = self.get_object_by_parsername_compare_genitiv("people", name)
        else:
            person = self.get_object_by_parsername("people", name)
        if person:
            return person
        elif not add:
            return None
        person_data.update({"name": name.strip().title(), "parser_names": name.strip()})
        if prefix:
            person_data["honorific_prefix"] = prefix
        response_data = self.parladata_api.people.set(person_data)
        return self.store_object(response_data, is_new=True)

    def add_person_parser_name(self, person: Person, parser_name: str) -> Person:
        updated_person = self.parladata_api.people.add_person_parser_name(
            person.id, parser_name
        )
        new_person = self.store_object(updated_person)
        del self.people[person.parser_name.lower()]
        return new_person

    def get_person_by_id(self, id: int) -> Person:
        if not self.people:
            self.load_data()
        return self.people_by_id.get(id, None)

    def get_prefix(self, name: str) -> tuple:
        prefix = re.findall("^[a-z]{0,4}\.", name)
        if prefix:
            prefix = prefix[0]
            return prefix, name.replace(prefix, "").strip().title()
        else:
            return None, name



import logging

from parladata_base_api.storages.utils import ParladataObject, Storage

logger = logging.getLogger("logger")


class Question(ParladataObject):
    keys = ["gov_id"]

    def __init__(
        self,
        gov_id: str,
        id: int,
        answer_timestamp: str,
        title: str,
        timestamp: str,
        is_new: bool,
        parladata_api,
    ) -> None:
        self.id = id
        self.gov_id = gov_id
        self.is_new = is_new
        self.answer_timestamp = answer_timestamp
        self.parladata_api = parladata_api
        self.title = title
        self.timestamp = timestamp

    def update_data(self, data: dict) -> dict:
        question = self.parladata_api.questions.patch(self.id, data)
        self.answer_timestamp = question["answer_timestamp"]
        return question

    def add_answer(self, data: dict) -> dict:
        data.update(question=self.id)
        return self.parladata_api.answers.set(data)


class QuestionStorage(Storage):
    def __init__(self, core_storage) -> None:
        super().__init__(core_storage)
        self.questions = {}
        self.storage = core_storage

    def load_data(self) -> None:
        if not self.questions:
            for question in self.parladata_api.questions.get_all(
                mandate=self.storage.mandate_id
            ):
                self.store_object(question, is_new=False)
            logger.info(f"laoded was {len(self.questions)} questions")

    def store_object(self, question: dict, is_new: bool) -> Question:
        temp_question = Question(
            gov_id=question["gov_id"],
            id=question["id"],
            answer_timestamp=question["answer_timestamp"],
            title=question["title"],
            timestamp=question["timestamp"],
            is_new=is_new,
            parladata_api=self.parladata_api,
        )
        self.questions[temp_question.get_key()] = temp_question
        return temp_question

    def get_or_add_object(self, data: dict) -> Question:
        if not self.questions:
            self.load_data()
        key = Question.get_key_from_dict(data)
        if key in self.questions.keys():
            return self.questions[key]
        else:
            data.update(mandate=self.storage.mandate_id)
            question = self.parladata_api.questions.set(data)
            new_question = self.store_object(question, is_new=True)
            return new_question

    def check_if_question_is_parsed(self, question: dict) -> bool:
        if not self.questions:
            self.load_data()
        key = Question.get_key_from_dict(question)
        return key in self.questions.keys()
